'''
Operations to handle different timezones.

'''

from typing import Union
from datetime import tzinfo, datetime

try:
    from zoneinfo import ZoneInfo as timezone
except ImportError:
    from backports.zoneinfo import ZoneInfo as timezone


def find_timezone(name: Union[str, tzinfo, timezone]) -> tzinfo:
    '''Retrieves the given timezone by name.'''
    # check if already converted
    if isinstance(name, (tzinfo, timezone)):
        return name

    # try to convert
    try:
        return timezone(name)
    except Exception:
        return None


def current_timezone() -> tzinfo:
    '''Retrieves the currently active timezone.'''
    # retrieve the current timezone name
    cur = datetime.now().astimezone().tzname()

    # special cases
    # note: this timezone is not available through the data (summer-time)
    if cur == "CEST":
        cur = "CET"

    return timezone(cur)



'''
All conversion function. This includes conversions between different datetimes.
'''

from datetime import datetime, date, tzinfo, time
from dateutil import parser
from typing import Any, Union
from logging import Logger

# try to import various dependend libraries
try:
    from zoneinfo import ZoneInfo as timezone
except ImportError:
    from backports.zoneinfo import ZoneInfo as timezone
try:
    import pandas as pd
    from pandas import Series, DataFrame
    from pandas.api.types import (
        is_datetime64_any_dtype as is_datetime
    )
    from pandas import to_datetime
except Exception:
    pd = None
    Series = None
    DataFrame = None
    is_datetime = lambda x: False
    to_datetime = lambda x: None
try:
    import numpy as np
    from numpy import array as nparray
except Exception:
    np = None
    nparray = None
from pytz import AmbiguousTimeError

from time_helper.const import DATE_FORMATS
from time_helper.timezone import current_timezone, find_timezone


def parse_time(time_str: str, format: str, timezone: Union[tzinfo, timezone, str]) -> datetime:
    '''Parses the given time based on the format and timezone (if provdied).

    Args:
        time_str: String value that should be parsed
        format: Format to parse the time from
        timezone: Timezone that should be applied (either str or actual timezone)

    Returns:
        (timzone-aware) datetime object
    '''
    # check the current timezone
    timezone = find_timezone(timezone)

    # update
    if isinstance(time_str, datetime):
        dt = time_str
    else:
        dt = datetime.strptime(time_str, format)
    if timezone is not None:
        dt = dt.replace(tzinfo=timezone)
    return dt


def unix_to_datetime(ts: Union[str, int, float, Any], tz: Union[timezone, str, Any] = None) -> datetime:
    '''Converts the given objects into a datetime.

    Args:
        ts: `int` or `long` that contains the timestamp
        tz: `pytz.timezone` that is used for localization

    Returns:
        `datetime` object that contains the time
    '''
    # check if should be parsed
    if isinstance(ts, str):
        try:
            ts = int(ts)
        except Exception:
            raise ValueError("Unable to convert object ({}) into a valid int or long item!".format(ts))
    # check if can be converted
    if isinstance(ts, int):
        # convert to datetime
        dt = datetime.utcfromtimestamp(ts)
        dt = dt.replace(tzinfo=timezone("UTC"))
        if tz is not None:
            dt = localize_datetime(dt, tz)
        else:
            print("WARNING: No timezone given for timestamp, infering 'UTC' as default!")
        return dt
    else:
        raise ValueError("Given object ({}) is not a valid int or long item!".format(ts))


def any_to_datetime(ts: Union[str, datetime, date, Any], logger: Logger = None, date_format: str = None) -> datetime:
    '''Generates a safe datetime from the input information.

    Args:
        ts: object to convert to datetime
        logger: Logging object to output infos
        date_format: Optional string with the date format to use (otherwise will try common ones)

    Returns:
        `datetime` object if converted or `None`
    '''
    dt = None

    # check if special case
    if ts is None or isinstance(ts, datetime):
        return ts

    # check if only date
    if isinstance(ts, date):
        dt = datetime.combine(ts, datetime.min.time())

    # convert from int or string
    if dt is None:
        try:
            dt = unix_to_datetime(ts)
        except Exception:
            pass

    # try relevant string formats
    if dt is None and isinstance(ts, str):
        # FEAT: improve list
        formats = DATE_FORMATS
        if date_format is not None:
            formats = [date_format] + formats

        try:
            dt = parser.isoparse(ts)
        except Exception:
            # check all formats
            for fmt in formats:
                try:
                    dt = datetime.strptime(ts, fmt)
                    if logger is not None:
                        logger.info(f"Date-Format '{fmt}' worked")
                except Exception:
                    if logger is not None:
                        logger.info(f"Date-Format '{fmt}' did not work")

    # check if only date
    if isinstance(dt, date):
        dt = datetime.combine(dt, datetime.min.time())

    # check for additional types
    if pd:
        if isinstance(dt, pd.Timestamp):
            dt = dt.to_pydatetime
        if dt == pd.NaT:
            return None
    if np:
        if isinstance(dt, np.datetime64):
            dt = dt.astype(datetime)
        if dt == np.nan:
            return None

    if dt is None:
        raise ValueError(f"Unable to parse datetime ({ts})")

    return dt


def convert_to_datetime(dt, baseline=None, remove_tz=False):
    '''Converts the given data to datetime.

    This might include conversions from date and time data-types

    Args:
        dt: Datetime, time or date object to convert
        baseline: datetime object to retrieve info from or None. Default is datetime.now()
        remove_tz: Defines if timezone information should be removed

    Returns:
        datetime object
    '''
    # check baseline
    if baseline is None:
        baseline = datetime.now()

    # check for types
    if isinstance(dt, datetime):
        pass
    elif isinstance(dt, time):
        dt = datetime(baseline.year, baseline.month, baseline.day, dt.hour, dt.minute, dt.second)
    elif isinstance(dt, date):
        dt = datetime(dt.year, dt.month, dt.day, 12, 0)
    else:
        raise ValueError(f"Given datetime data has unkown type ({type(dt)}")

    # check for removal
    if remove_tz:
        dt = dt.replace(tzinfo=None)

    return dt


def localize_datetime(dt: datetime, tz: Union[Any, str, timezone] = None) -> datetime:
    '''Localizes a datetime to the current timezone.

    Args:
        dt (datetime): Datetime to make aware
        tz (str, timezone): Timezone (either directly or name of the timezone)
    '''
    # check if None
    if dt is None:
        return None
    if tz is None:
        tz = current_timezone()

    # update the timezone
    if isinstance(tz, str):
        tz = timezone(tz)

    # check if timezone should be added or converted
    if dt.tzinfo is None:
        return dt.replace(tzinfo=tz)
    else:
        return dt.astimezone(tz)


def make_aware(dt: Union[datetime, str, Any], tz: Union[str, timezone, Any] = None, force_convert: bool = True, col: str = None) -> Union[datetime, Any]:
    '''Checks if the current datetime is aware, otherwise make aware.

    Args:
        dt (datetime): Datetime to convert
        tz (str): Name of the timezone to convert to
        force_convert (bool): Defines if the timezone should be converted if there is already a timezone present
        col (str): Column used if the input is pandas Dataframe

    Returns:
        Updated datetime (or pandas object)
    '''
    # ensure that data is not none
    if dt is None:
        return None

    # check for pandas
    is_pandas = False
    try:
        is_pandas = isinstance(dt, (Series, DataFrame))
    except Exception:
        pass

    if is_pandas:
        return make_aware_pandas(dt, col, tz=tz)

    # make sure dt is datetime
    dt = any_to_datetime(dt)
    if dt is None:
        return None

    # check if already aware
    if dt.tzinfo is not None and (tz is None or force_convert is False):
        return dt

    # check for local timezone (if none provided)
    if tz is None:
        tz = current_timezone()

    # return localized datetime
    return localize_datetime(dt, tz)


def make_aware_pandas(df: Union[Series, DataFrame], col: str, format=None, tz=None) -> Union[Series, DataFrame]:
    '''This will make the pandas column datetime aware in the specified timezone.

    Defaults the data to the current timezone.

    Args:
        df: DataFrame to convert
        col: name of the column to convert
        format: Default format to try
        timezone: default timezone to convert to

    Returns:
        Updated DataFrame
    '''
    # check if pandas is install
    if not Series or not DataFrame:
        raise ImportError("Pandas Library is not installed")

    # safty checks
    if df is None:
        return None
    if col is None:
        raise ValueError("Expected column name, but got None")
    if col not in df:
        raise RuntimeError(f"The specified column {col} is not available in the dataframe: {df.columns}")

    # make sure the data is unaware
    if not is_datetime(df[col]):
        # generate format list
        formats = []
        if format is not None:
            formats.append(format)
        formats.append(None)
        formats += DATE_FORMATS

        # check all formats
        for fmt in formats:
            try:
                df[col] = to_datetime(df[col], format=fmt)
            except Exception:
                pass

    # TODO: update timezone ensurances
    # ensure timezone
    if not hasattr(df[col].iloc[0], 'tzinfo') or not df[col].iloc[0].tzinfo:
        cur_tz = current_timezone().key
        try:
            df[col] = df[col].dt.tz_localize(cur_tz)
        except AmbiguousTimeError:
            infer_dst = nparray([False] * df.shape[0])
            df[col] = df[col].dt.tz_localize(cur_tz, ambiguous=infer_dst)
    if tz is not None:
        # convert to string (as pandas does not support ZoneInfo)
        if isinstance(tz, timezone):
            tz = tz.key
        df.loc[:, col] = df[col].dt.tz_convert(tz)

    return df


def make_unaware(dt: Union[datetime, Any], tz: Union[str, tzinfo, timezone] = "UTC") -> datetime:
    '''Makes the given timezone unaware in a default timezone.

    Args:
        dt: Datetime to convert
        tz: Timezone to convert to as basetime

    Returns:
        datetime object without timezone info
    '''
    # check against None values
    if dt is None:
        return None

    # ensure the datetime is safe
    dt = any_to_datetime(dt)

    # convert the timezone
    return localize_datetime(dt, tz).replace(tzinfo=None)




# define the version
__version__ = "0.1.3"

import logging

try:
    from . import const
    from .timezone import (
        find_timezone,
        current_timezone
    )

    from .convert import (
        parse_time,
        unix_to_datetime,
        any_to_datetime,
        convert_to_datetime,
        localize_datetime,
        make_aware,
        make_unaware
    )

    from .ops import (
        has_timezone,
        time_diff,
        round_time
    )

    from .range import (
        time_to_interval,
        create_intervals
    )

    from .wrapper import DateTimeWrapper
except Exception:
    logging.error("Not all dependencies installed")




'''
Contains a range of operations on different datetimes.
'''

from datetime import datetime, date, timedelta
from typing import Union

try:
    from pandas import Series, DataFrame
    from pandas.api.types import (
        is_datetime64_any_dtype as is_datetime
    )
except Exception:
    Series = None
    DataFrame = None
    is_datetime = lambda x: False

from time_helper.convert import make_aware, any_to_datetime, localize_datetime


def has_timezone(df: Union[Series, DataFrame], col: str = None) -> bool:
    '''Checks if a given pandas object has a timezone.

    Args:

    '''
    # perform security
    if df is None:
        raise ValueError("Expected a dataframe but got None")
    if isinstance(df, DataFrame):
        if col is None:
            raise ValueError("Expected a column name, but got None")
        if col not in df.columns:
            raise ValueError(f"The provided column {col} is not in the dataframe {df.columns}")
        if not is_datetime(df[col]):
            raise ValueError("Specified column is not a datetime object!")
    elif not is_datetime(df):
        raise ValueError("Provided series is not a datetime object!")
    if len(df) == 0:
        raise ValueError("The Dataframe is empty")

    # perform checks
    df_col = df if isinstance(df, Series) else df[col]
    obj = df_col.iloc[0]
    if not hasattr(obj, 'tzinfo'):
        return False
    if obj.tzinfo is None:
        return False
    return True


def time_diff(dt1: datetime, dt2: datetime, tz=None) -> timedelta:
    '''Allows to compute the difference between datetimes of different timezones.

    Calculates dt1 - dt2

    Args:
        dt1 (datetime): First datetime in the diff
        dt2 (datetime): Second datetime in the diff
        tz (str): Name of the default timezone to use

    Returns:
        timedelta
    '''
    # check for unaware
    if dt1.tzinfo is None and dt2.tzinfo is None:
        return dt1 - dt2

    # update timezones
    if dt1.tzinfo is None:
        dt1 = make_aware(dt1, tz, force_convert=False)
    if dt2.tzinfo is None:
        dt2 = make_aware(dt2, tz, force_convert=False)

    # bring both datetimes to same timezone
    dt1 = localize_datetime(dt1, 'UTC')
    dt2 = localize_datetime(dt2, 'UTC')

    # calculate
    return dt1 - dt2


def round_time(dt: datetime, freq: str = "D", max_out: bool = False) -> datetime:
    '''Rounds the given timestamp to the start or end of the day.

    Args:
        timestamp (datetime): Datetime that should be rounded
        freq (str): Frequency to round it to (options are S, M, H, D, W, m, Y)
        max_out (bool): Defines if remaining attributes should be zeroed (False) or maxed out (True)

    Returns:
        Updated datetime
    '''
    # check if value is provided
    if not dt:
        return None

    # ensure value is a datetime
    if not isinstance(dt, datetime):
        try:
            dt = any_to_datetime(dt)
        except Exception:
            raise ValueError(f"Could not parse Timestamp: {dt}")

    # check for date
    if not isinstance(dt, datetime) and isinstance(dt, date) and freq in ["H", "M", "S"]:
        raise ValueError("Got a date, but frequency requires datetime")

    # check for week special case
    if freq == "W":
        day = dt.weekday()
        dt = dt + timedelta(days=6 - day) if max_out else dt - timedelta(days=day)

    # update time to get most out of day
    items = {"microsecond": 999999 if max_out else 0}
    if freq not in ["S"]:
        items["second"] = 59 if max_out else 0
    if freq not in ["S", "M"]:
        items["minute"] = 59 if max_out else 0
    if freq not in ["S", "M", "H"]:
        items["hour"] = 23 if max_out else 0
    if freq in ["Y", "m"]:
        day_in_month = (dt.replace(month=dt.month % 12 + 1, day=1) - timedelta(days=1)).day
        items["day"] = day_in_month if max_out else 1
    if freq in ["Y"]:
        items["month"] = 12 if max_out else 1
        items["day"] = 31 if max_out else items["day"]
    dt = dt.replace(**items)

    return dt



'''
Wrapper Class for different time objects (including dates and datetimes).

This makes many of the functions first class citizens and can easily expose the datetime
'''

from datetime import datetime
from typing import Any

from time_helper.convert import any_to_datetime
from time_helper.ops import time_diff


class DateTimeWrapper():
    def __init__(self, dt) -> None:
        self.dt = any_to_datetime(dt)

    def __call__(self, *args: Any, **kwds: Any) -> datetime:
        if not args and not kwds:
            return self.dt
        # TODO: call action on the internal datetime object

    # TODO: overload operators
    def __sub__(self, other):
        # make sure to convert
        if not isinstance(other, DateTimeWrapper):
            other = DateTimeWrapper(other)

        # ensure to compute
        return time_diff(self.dt, other.dt)




import os
from setuptools import setup
from setuptools import find_packages
from time_helper import __version__
import pathlib


__status__      = "Package"
__copyright__   = "Copyright 2022"
__license__     = "MIT License"
__author__      = "Felix Geilert"


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'readme.md'), encoding='utf-8') as f:
    long_description = f.read()


def versions_in_requirements(file):
    lines = file.read().splitlines()
    versions = [
        line
        for line in lines
        if not line.isspace() and "--" not in line
    ]
    return list(versions)


HERE = pathlib.Path(__file__).parent
with open(HERE / "requirements.txt") as f:
    required_list = versions_in_requirements(f)

setup(
    name='time-helper',
    version=__version__,
    description='Various Helper Tools to handle different time data',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='time handling',
    url='https://github.com/felixnext/python-time-helper',
    download_url='https://github.com/felixnext/python-time-helper/releases/tag/v0.1.0-alpha',
    author='Felix Geilert',
    license='MIT License',
    packages=find_packages(include=['time_helper', 'time_helper.*']),
    install_requires=required_list,
    setup_requires=['pytest-runner', 'flake8'],
    tests_require=['pytest'],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)
